import pandas as pds

class custom(object):
    """
    Class holds a list of functions and arguments to be applied when load called.
    """

    def __init__(self):
        # create a list object to store functions
        self._functions = []
        # type of function stored (add/modify/pass on data)
        self._kind = []
        # arguments to functions
        self._args = []
        # keyword arguments to functions
        self._kwargs = []

    def add(self, function, kind='add', at_pos='end',*args, **kwargs):
        """ Add a function to custom processing queue. 
        
        Custom functions are applied automatically to associated
        pysat instrument whenever instrument.load command called.

        Inputs:
            function - name of function or function object
            
            kind - type of custom function
                add : Adds data returned from function to instrument object. 
                      A copy of pysat instrument object supplied to routine. 
                 Allowed returns:
                    {'data':pandas Series/DataFrame/numpy array/list,
                    'units':string/list of strings, 
                    'long_name':string/list of strings,
                    'name':string/list of strings (iff data isarray or list)}
                        
                    or, pandas DataFrame, names of columns are used
                    or, pandas Series, .name required 
                    or, (string/list of strings, numpy array/list of arrays)
                modify : pysat instrument object supplied to routine. Any and all
                         changes to object are retained.
                pass : A copy of pysat object is passed to function. No 
                       data is accepted from return.
                       
            at_pos - insert at position. (default, insert at end).                      
        """

        if isinstance(function, str):
            # convert string to function object
            function=eval(function)

        if (at_pos == 'end') | (at_pos == len(self._functions)):
            # store function object
            self._functions.append(function)
            self._args.append(args)
            self._kwargs.append(kwargs)
            self._kind.append(kind.lower())
        elif at_pos < len(self._functions):
            # user picked a specific location to insert
            self._functions.insert(at_pos, function)
            self._args.insert(at_pos, args)
            self._kwargs.insert(at_pos, kwargs)
            self._kind.insert(at_pos, kind)
        else:
            raise TypeError('Must enter an index between 0 and %i' % len(self._functions))

    def _apply_all(self, sat):
        '''
        Apply all of the custom functions to the satellite data object.
        '''
        if len(self._functions) > 0:
            for func, arg, kwarg, kind in zip(self._functions, self._args, 
                                                self._kwargs, self._kind):
                if len(sat.data) > 0:     
                    if kind == 'add':
                        # apply custom functions that add data to the instrument object
                        tempd = sat.copy()
                        newData = func(tempd, *arg, **kwarg)
                        del tempd
                        
                        # not fancy below, but it works
                        if isinstance(newData,dict):
                            if isinstance(newData['data'], pds.DataFrame):
                                sat[newData['data'].columns] = newData
                            elif isinstance(newData['data'], pds.Series):
                                if newData['data'].name is not None:
                                    sat[newData['data'].name] = newData
                                elif 'name' in newData.keys():
                                    name = newData.pop('name')
                                    sat[name] = newData
                                else:
                                    raise ValueError('Must assign a name to Series'+
                                            ' or return a "name" in dictionary.')
                            elif hasattr(newData['data'], '__iter__'):
                                if 'name' in newData.keys():
                                    name = newData.pop('name')
                                    sat[name] = newData
                                else:
                                    raise ValueError('Must include "name" in returned dictionary.')
                                    
                        elif isinstance(newData, pds.DataFrame):
                            sat[newData.columns] = newData
                        elif isinstance(newData, pds.Series):
                            sat[newData.name] = newData                            
                        elif hasattr(newData, '__iter__'):
			    # falling back to older behavior 
			    # unpack tuple/list that was returned 
                            newName = newData[0]
                            newData = newData[1]
                            if len(newData)>0:
                                # doesn't really check ensure data, there could
                                # be multiple empty arrays returned, [[],[]]
                                if isinstance(newName, str):
                                    # one item to add
                                    sat[newName] = newData
                                else:    		
                                    # multiple items
                                    for name, data in zip(newName, newData):
                                        if len(data)>0:        
                                            # fixes up the incomplete check from before
                                            sat[name] = data
                        else:
                            raise ValueError("kernel doesn't know what to do with returned data.")
                    # modifying loaded data
                    if kind == 'modify':
                        t = func(sat,*arg,**kwarg)
			if t is not None:
			    raise ValueError('Modify functions should not return any information via return. Information may only be propagated back by modifying supplied pysat object.')	
                    # pass function (function runs, no data allowed back)
                    if kind == 'pass':
                        tempd = sat.copy()
                        t = func(tempd,*arg,**kwarg)
                        del tempd
			if t is not None:
			    raise ValueError('Pass functions should not return any information via return.')	

    def clear(self):
        """Clear custom function list."""
        self._functions=[]
        self._args=[]
        self._kwargs=[]
        self._kind=[]

######################################################
##### END CUSTOM CLASS ##############################
####################################################



import pysat
import pandas as pds
import numpy as np
from nose.tools import assert_raises, raises
import nose.tools

class TestBasics:
    def setup(self):
	'''Runs before every method to create a clean testing setup.'''
        self.testInst = pysat.Instrument('pysat','testing', '10', 'clean')

    def teardown(self):
        '''Runs after every method to clean up previous testing.'''
        del self.testInst

    @raises(ValueError)        
    def test_single_modifying_custom_function(self):
	'''Test if custom function works correctly. Modify function that returns pandas object. Modify function returns an object which will produce an Error.'''
        def custom1(inst):
            inst.data['doubleMLT'] = 2.*inst.data.mlt
            return 5.*inst.data
	
        self.testInst.custom.add(custom1, 'modify')  
        self.testInst.load(2009,1)

    def test_single_adding_custom_function(self):
	'''Test if custom function works correctly. Add function that returns pandas object.'''
        def custom1(inst):
            d = 2.*inst.data.mlt
            d.name='doubleMLT'
            return d

        self.testInst.custom.add(custom1, 'add')  
        self.testInst.load(2009,1)
        ans = (self.testInst.data['doubleMLT'].values == 2.*self.testInst.data.mlt.values).all()
        assert ans

    def test_single_adding_custom_function_wrong_times(self):
	'''Only the data at the correct time should be accepted, otherwise nan'''
        def custom1(inst):
            d = 2.*inst.data.mlt
            d.name='doubleMLT'
            d.index += pds.DateOffset(microseconds=10)
            return d

        self.testInst.custom.add(custom1, 'add')  
        self.testInst.load(2009,1)
        ans = (self.testInst.data['doubleMLT'].isnull()).all()
        assert ans

    def test_single_adding_custom_function_that_modifies_passed_data(self):
	'''Test if custom function works correctly. Add function that returns pandas object but modifies passed satellite object. 
		Changes to passed object should not propagate back.'''
        def custom1(inst):
            inst.data['doubleMLT'] = 2.*inst.data.mlt
            inst.data.mlt=0.
            return inst.data.doubleMLT

        self.testInst.custom.add(custom1, 'add')  
        self.testInst.load(2009,1)
        ans = (self.testInst.data['doubleMLT'] == 2.*self.testInst.data.mlt).all()
        assert ans

    def test_add_function_tuple_return_style(self):
	'''Test if custom function works correctly. Add function that returns name and numpy array.'''
        def custom1(inst):
            return ('doubleMLT',2.*inst.data.mlt.values)
        self.testInst.custom.add(custom1, 'add')  
        self.testInst.load(2009,1)
        ans = (self.testInst.data['doubleMLT'] == 2.*self.testInst.data.mlt).all()
        assert ans
        
    def test_add_multiple_custom_functions_tuple_return_style(self):
	'''Test if multiple custom functions that add data work correctly. Add function that returns name and numpy array.'''
        def custom1(inst):
            return (['doubleMLT', 'tripleMLT'],[2.*inst.data.mlt.values, 3.*inst.data.mlt.values])
        self.testInst.custom.add(custom1, 'add')  
        self.testInst.load(2009,1)
        ans = (((self.testInst.data['doubleMLT'] == 2.*self.testInst.data.mlt).all()) & ((self.testInst.data['tripleMLT'] == 3.*self.testInst.data.mlt).all()))
        assert ans

    @raises(ValueError)
    def test_add_function_tuple_return_style_too_few_elements(self):
	'''Test if custom function works correctly. Add function that returns name and numpy array.'''
        def custom1(inst):
            return ('doubleMLT',2.*inst.data.mlt.values[0:-5])
        self.testInst.custom.add(custom1, 'add')  
        self.testInst.load(2009,1)

    @raises(ValueError)
    def test_add_function_tuple_return_style_too_many_elements(self):
	'''Test if custom function works correctly. Add function that returns name and numpy array.'''
        def custom1(inst):
            return ('doubleMLT',np.arange(2.*len(inst.data.mlt)))
        self.testInst.custom.add(custom1, 'add')  
        self.testInst.load(2009,1)
                                                        
    def test_add_dataframe(self):
        def custom1(inst):
            out = pds.DataFrame({'doubleMLT':inst.data.mlt*2, 
                                'tripleMLT':inst.data.mlt*3}, 
                                index=inst.data.index)
            return out
        self.testInst.custom.add(custom1, 'add')
        self.testInst.load(2009,1)
        ans = (((self.testInst.data['doubleMLT'] == 2.*self.testInst.data.mlt).all()) & ((self.testInst.data['tripleMLT'] == 3.*self.testInst.data.mlt).all()))
        assert ans

    def test_add_dataframe_w_meta(self):
        def custom1(inst):
            out = pds.DataFrame({'doubleMLT':inst.data.mlt*2, 
                                'tripleMLT':inst.data.mlt*3}, 
                                index=inst.data.index)
            return {'data':out, 'long_name':['doubleMLTlong', 'tripleMLTlong'],
                    'units':['hours1', 'hours2']}
        self.testInst.custom.add(custom1, 'add')
        self.testInst.load(2009,1)
        ans1 = self.testInst.meta['doubleMLT'].units == 'hours1'
        ans2 = self.testInst.meta['doubleMLT'].long_name == 'doubleMLTlong'
        ans3 = self.testInst.meta['tripleMLT'].units == 'hours2'        
        ans4 = self.testInst.meta['tripleMLT'].long_name == 'tripleMLTlong'
        ans5 = (self.testInst['doubleMLT'] == 2.*self.testInst.data.mlt).all()
        ans6 = (self.testInst['tripleMLT'] == 3.*self.testInst.data.mlt).all()
        assert ans1 & ans2 & ans3 & ans4 & ans5 & ans6
        
    def test_add_series_w_meta(self):
        def custom1(inst):
            out = pds.Series(inst.data.mlt*2, 
                                index=inst.data.index)
            out.name = 'doubleMLT'
            return {'data':out, 'long_name':'doubleMLTlong',
                    'units':'hours1'}
        self.testInst.custom.add(custom1, 'add')
        self.testInst.load(2009,1)
        ans1 = self.testInst.meta['doubleMLT'].units == 'hours1'
        ans2 = self.testInst.meta['doubleMLT'].long_name == 'doubleMLTlong'
        ans3 = (self.testInst['doubleMLT'] == 2.*self.testInst.data.mlt).all()
        assert ans1 & ans2 & ans3

    def test_add_series_w_meta_missing_long_name(self):
        def custom1(inst):
            out = pds.Series(2.*inst.data.mlt.values, 
                                index=inst.data.index)
            out.name = 'doubleMLT'
            return {'data':out, 
                    'units':'hours1'}
        self.testInst.custom.add(custom1, 'add')
        self.testInst.load(2009,1)
        ans1 = self.testInst.meta['doubleMLT'].units == 'hours1'
        ans2 = self.testInst.meta['doubleMLT'].long_name == 'doubleMLT'
        ans3 = (self.testInst['doubleMLT'] == 2.*self.testInst.data.mlt).all()
        assert ans1 & ans2 & ans3        
        
    def test_add_series_w_meta_name_in_dict(self):
        def custom1(inst):
            out = pds.Series(2.*inst.data.mlt.values, 
                                index=inst.data.index)
            return {'data':out, 'long_name':'doubleMLTlong',
                    'units':'hours1', 'name':'doubleMLT'}
        self.testInst.custom.add(custom1, 'add')
        self.testInst.load(2009,1)
        ans1 = self.testInst.meta['doubleMLT'].units == 'hours1'
        ans2 = self.testInst.meta['doubleMLT'].long_name == 'doubleMLTlong'
        ans3 = (self.testInst['doubleMLT'] == 2.*self.testInst.data.mlt).all()
        assert ans1 & ans2 & ans3
        
    @raises(ValueError)    
    def test_add_series_w_meta_no_name(self):
        def custom1(inst):
            out = pds.Series({'doubleMLT':inst.data.mlt*2}, 
                                index=inst.data.index)
            #out.name = 'doubleMLT'
            return {'data':out, 'long_name':'doubleMLTlong',
                    'units':'hours1'}
        self.testInst.custom.add(custom1, 'add')
        self.testInst.load(2009,1)   

    def test_add_numpy_array_w_meta_name_in_dict(self):
        def custom1(inst):
            out = (inst.data.mlt*2).values
            return {'data':out, 'long_name':'doubleMLTlong',
                    'units':'hours1', 'name':'doubleMLT'}
        self.testInst.custom.add(custom1, 'add')
        self.testInst.load(2009,1)
        ans1 = self.testInst.meta['doubleMLT'].units == 'hours1'
        ans2 = self.testInst.meta['doubleMLT'].long_name == 'doubleMLTlong'
        ans3 = (self.testInst['doubleMLT'] == 2.*self.testInst.data.mlt).all()
        assert ans1 & ans2 & ans3

    @raises(ValueError)  
    def test_add_numpy_array_w_meta_no_name_in_dict(self):
        def custom1(inst):
            out = (inst.data.mlt*2).values
            return {'data':out, 'long_name':'doubleMLTlong',
                    'units':'hours1'}
        self.testInst.custom.add(custom1, 'add')
        self.testInst.load(2009,1)


    def test_add_list_w_meta_name_in_dict(self):
        def custom1(inst):
            out = (inst.data.mlt*2).tolist()
            return {'data':out, 'long_name':'doubleMLTlong',
                    'units':'hours1', 'name':'doubleMLT'}
        self.testInst.custom.add(custom1, 'add')
        self.testInst.load(2009,1)
        ans1 = self.testInst.meta['doubleMLT'].units == 'hours1'
        ans2 = self.testInst.meta['doubleMLT'].long_name == 'doubleMLTlong'
        ans3 = (self.testInst['doubleMLT'] == 2.*self.testInst.data.mlt).all()
        assert ans1 & ans2 * ans3

    @raises(ValueError)  
    def test_add_list_w_meta_no_name_in_dict(self):
        def custom1(inst):
            out = (inst.data.mlt*2).tolist()
            return {'data':out, 'long_name':'doubleMLTlong',
                    'units':'hours1'}
        self.testInst.custom.add(custom1, 'add')
        self.testInst.load(2009,1)
    


# -*- coding: utf-8 -*-
"""
Created on Mon Feb 11 11:17:56 2013

@author: rstoneba
"""
import glob
import os
import netCDF4
import pandas as pds
import numpy as np
import pysat


def list_files(tag=None, data_path=None):
    """Return a Pandas Series of every file for chosen satellite data"""
    import sys
    #if tag == 'ionprf':
    #    # from_os constructor currently doesn't work because of the variable 
    #    # filename components at the end of each string.....
    #    #return pysat.Files.from_os(dir_path=os.path.join('cosmic', 'ionprf'),
    #        #format_str='*/ionPrf_*.{year:04d}.{day:03d}.{hour:02d}.{min:02d}*_nc')          
    print 'Building a list of COSMIC files, which can possibly take time. ~1s per 100K files'
    sys.stdout.flush()

    # number of files may be large, written with this in mind
    # only select file that are the cosmic data files and end with _nc
    cosmicFiles = glob.glob(os.path.join(data_path, '*/*_nc'))
    # need to get date and time from filename to generate index
    num = len(cosmicFiles) 
    if num != 0:
        print 'Estimated time:', num*1.E-5,'seconds'
        sys.stdout.flush()
        # preallocate lists
        year=[None]*num; days=[None]*num; hours=[None]*num; 
        minutes=[None]*num; microseconds=[None]*num;
        for i,f in enumerate(cosmicFiles):
            f2 = f.split('.')
            year[i]=f2[2]
            days[i]=f2[3]
            hours[i]=f2[4]
            minutes[i]=f2[5]
            microseconds[i]=i
    
        year=np.array(year).astype(int)
        days=np.array(days).astype(int)
        uts=np.array(hours).astype(int)*3600+np.array(minutes).astype(int)*60
        # adding microseconds to ensure each time is unique, not allowed to pass 1.E-3 s
        uts+=np.mod(np.array(microseconds).astype(int)*1.E-6, 1.E-3)
        index = pysat.utils.create_datetime_index(year=year, doy=days, uts=uts)
        file_list = pds.Series(cosmicFiles, index=index)
        return file_list
    else:
        print 'Found no files, check your path or download them.'
        return pds.Series(None)
        

def load(cosmicFiles, tag=None):
    """
    cosmic data load routine, called by pysat
    """   
    num = len(cosmicFiles)
    # make sure there are files to read
    if num != 0:
        # call separate load_files routine, segemented for possible
        # multiprocessor load, not included and only benefits about 20%
        output = pds.DataFrame(load_files(cosmicFiles, tag=tag))
        output.index = pysat.utils.create_datetime_index(year=output.year, 
                month=output.month, doy=output.day, 
                uts=output.hour*3600.+output.minute*60.+output.second)
        # make sure UTS strictly increasing
	output.sort(inplace=True)	
	# use the first available file to pick out meta information
	meta = pysat.Meta()
	ind = 0 
	repeat = True
	while repeat:
            try:
                data = netCDF4.Dataset(cosmicFiles[ind]) 
                ncattrsList = data.ncattrs()
                for d in ncattrsList:
                    meta[d] = {'units':'', 'long_name':d}
                keys = data.variables.keys()
                for key in keys:
                    meta[key] = {'units':data.variables[key].units, 
                                'long_name':data.variables[key].long_name}  
                repeat = False                  
            except RuntimeError:
                # file was empty, try the next one by incrementing ind
                ind+=1
                                    
        return output, meta
    else:
        # no data
        return pds.DataFrame(None), pysat.Meta()



# seperate routine for doing actual loading. This was broken off from main load
# becuase I was playing around with multiprocessor loading
# yielded about 20% improvement in execution time
def load_files(files, tag=None):
    '''Loads a list of COSMIC data files, supplied by user.
    
    Returns a list of dicts, a dict for each file.
    '''
       
    output = [None]*len(files)
    drop_idx = []
    for (i,file) in enumerate(files):
        try:
            data = netCDF4.Dataset(file)    
            # build up dictionary will all ncattrs
            new = {} 
            # get list of file attributes
            ncattrsList = data.ncattrs()
            for d in ncattrsList:
                new[d] = data.getncattr(d)            
            # load all of the variables in the netCDF
            loadedVars={}
            keys = data.variables.keys()
            for key in keys:
                loadedVars[key] = data.variables[key][:]               
            new['profiles'] = pds.DataFrame(loadedVars)
            if tag == 'ionprf':
                new['profiles'].index = new['profiles']['MSL_alt']
            output[i] = new   
            data.close()
        except RuntimeError:
            # some of the S4 files have zero bytes, which causes a read error
            # this stores the index of these zero byte files so I can drop 
            # the Nones the gappy file leaves behind
            drop_idx.append(i)

    # drop anything that came from the zero byte files
    drop_idx.reverse()
    for i in drop_idx:
        del output[i]
    return output


def clean(self):
    
    if self.tag == 'ionprf':
        # ionosphere density profiles
        if self.clean_level == 'clean':
            # try and make sure all data is good
            # filter out profiles where source provider processing doesn't get max dens and max dens alt
            self.data = self.data[( (self['edmaxalt'] != -999.) & (self['edmax'] != -999.) )]
            # make sure edmaxalt in "reasonable" range
            self.data = self.data[(self.data.edmaxalt >= 175.) & (self.data.edmaxalt <= 475.)]
            # filter densities when negative
            for i, profile in enumerate(self['profiles']):
                # take out all densities below the highest altitude negative dens below 325
                idx, = np.where((profile.ELEC_dens < 0) & (profile.index <= 325))	
                if len(idx) > 0:
                    profile.iloc[0:idx[-1]+1] = np.nan
                # take out all densities above the lowest altitude negative dens above 325
                idx, = np.where((profile.ELEC_dens < 0) & (profile.index > 325))	
                if len(idx) > 0:
                    profile.iloc[idx[0]:] = np.nan
                                    
                # do an altitude density gradient check to reduce number of cycle slips 
                densDiff = profile.ELEC_dens.diff()
                altDiff = profile.MSL_alt.diff()
                normGrad = ( densDiff/(altDiff*profile.ELEC_dens) ).abs()
           	idx, = np.where((normGrad > 1.) & normGrad.notnull())
           	if len(idx) > 0:
           	    self[i,'edmaxalt'] = np.nan
           	    self[i,'edmax'] = np.nan
           	    self[i,'edmaxlat'] = np.nan
           	    profile['ELEC_dens'] *= np.nan
           	        #self.data['profiles'][i]['ELEC_dens'] *= np.nan
    
       	# filter out any measurements where things have been set to NaN    
       	self.data = self.data[self.data.edmaxalt.notnull()]  
           	  
    elif self.tag == 'scnlvl1':
        # scintillation files
        if self.clean_level == 'clean':
            # try and make sure all data is good       
            # filter out profiles where source provider processing doesn't 
            # work 
            self.data = self.data[( (self['alttp_s4max'] != -999.) & (self['s4max9sec'] != -999.) )]

    return

def download(date_array, tag, data_path=None, user=None, password=None):
    import ftplib
    import urllib2
    import base64
    import os
    import tarfile
    import shutil

    if tag == 'ionprf':
        sub_dir = 'ionPrf'
    elif tag == 'sonprf':
        sub_dir = 'sonPrf'
    elif tag == 'wetprf':
        sub_dir = 'wetPrf'
    elif tag == 'atmPrf':
        sub_dir = 'atmPrf'
    else:
        raise ValueError('Unknown cosmic_gps tag')
        
    for date in date_array:
        yr,doy = pysat.utils.getyrdoy(date)
        yrdoystr = '{year:04d}.{doy:03d}'.format(year=yr, doy=doy)
        dwnld="http://cdaac-www.cosmic.ucar.edu/cdaac/rest/tarservice/data/cosmic2013/"
        dwnld = dwnld+sub_dir+'/{year:04d}.{doy:03d}'.format(year=yr, doy=doy)
        req=urllib2.Request(dwnld)
        base64str=base64.encodestring('%s:%s' % (user, password)).replace('\n', '')
        req.add_header("Authorization", "Basic %s" % base64str)   
        result=urllib2.urlopen(req)
        fname=os.path.join(data_path,'cosmic_'+sub_dir+'_'+ yrdoystr+'.tar' )
        with open(fname, "wb") as local_file:
            local_file.write(result.read())
            local_file.close()
            # uncompress files
            tar = tarfile.open(fname)
            tar.extractall(path=data_path)
            tar.close()
            # move files
            ext_dir = os.path.join(data_path,'cosmic2013',sub_dir,yrdoystr) 
            shutil.move(ext_dir, os.path.join(data_path,yrdoystr))

    return

    
    
    
    
    ## mean altitude profiles over bin size, make a pandas Series for each
    #altBin = 3
    #roundMSL_alt = np.round(loadedVars['MSL_alt']/altBin)*altBin
    #profiles = pds.DataFrame(loadedVars, index=roundMSL_alt)
    #profiles = profiles.groupby(profiles.index.values).mean()
    #del loadedVars 


# -*- coding: utf-8 -*-
import string
import pandas as pds
import numpy as np
import os
#from types import ModuleType
import copy
import sys


import _custom
import _files
import _orbits
import _meta
import utils
from pysat import data_dir #import pysat.data_dir as data_dir

# main class for users
class Instrument(object):


    def __init__(self, platform=None, name=None, tag=None, clean_level='clean', 
                query_files=False, pad=False,
                orbit_index=None, orbit_type=None, orbit_period=None,  
                inst_module=None, *arg, **kwargs):
        '''
        Class for analyzing, modifying, and managing satellite data.
        
        platform: String for instrument platform/satellite.
	name : String indicating name of instrument. 
	tag : String identifying particular aspect of the instrument.
	   pysat attempts to load corresponding instrument module platform_name.
	inst_module : Provide instrument module directly (takes precedence over name)
	clean_level : String passed to clean function indicating level of quality.
	pad : Boolean indicating whether padding should be added (using instrument data) to begining and end of data
	       for time-series processing. Extra data is removed after applying all custom functions. Pad=True requires
                length of time for padding (seconds or minutes or hours). Maximum of 1-day.
	orbit_index : String indicating which data will be used to determine breaks in orbit, if requested.
	orbit_type : String indicating type of orbit, equatorial (default) or polar.
	orbit_period : datetime.timedelta object containing length of orbital period.
	query_files: Boolean indicating whether the filesystem should be queried
	               or if the stored list of files is sufficient. (set True if new files)
	
	'''

        if inst_module is None:
            if isinstance(platform, str) and isinstance(name, str):              
                self.platform = platform.lower() 
                self.name = name.lower() 
                self._assign_funcs(by_name=True)
            elif (platform is None) and (name is None):
                # creating "empty" Instrument object with this path
                self.name = ''
                self.platform = ''
                self._assign_funcs()                
            else:
                raise ValueError('Inputs platform and name must both be strings, or both None.')                        
        else:
            # user expected to have provided module
            try:
                self.name = inst_module.name.lower()
                self.platform = inst_module.platform.lower()
            except AttributeError:
 	        raise AttributeError(string.join(('A name and platform attribute for the ',
                    'instrument is required if supplying routine module directly.')))
            self._assign_funcs(inst_module=inst_module)
            
        self.tag = tag.lower() if tag is not None else ''    
        self.clean_level = clean_level.lower() if clean_level is not None else 'none'

        self.data = pds.DataFrame(None)
        self.meta = _meta.Meta()
        # function processing class, processes data on load
        self.custom = _custom.custom()
        # create arrays to store data around loaded day
        # enables padding across day breaks with minimal loads
        self._next_data = pds.DataFrame(None)
	self._next_data_track = []
        self._prev_data = pds.DataFrame(None)
        self._prev_data_track = []
        self._curr_data = pds.DataFrame(None)
        # arguments for padding
        self.pad = pad
        self.padkws = kwargs	

        # load file list function, which returns dict of files
        # as well as data start and end dates
        self.files = _files.Files(self)
        if query_files:
            self.files.refresh()
        # set bounds for iteration based upon data properties
        self.bounds = (None, None)
        self.date = None
        self._fid = None
        self.yr = None
        self.doy = None
        self._load_by_date = False
        # attach seasonal methods for convenience        
        #self.ssnl = ssnl
        # initiliaze orbit support
        self.orbits = _orbits.orbits(self, index=orbit_index, kind=orbit_type,
                                    period=orbit_period)
        # run instrument init function, a basic pass function is used
        # if user doesn't supply the init function
        self._init_rtn(self)
        
                
    def __getitem__(self, key): 
        """
        Convenience notation for accessing data in inst.data.name using
        inst['name'].
        
        Slicing is also allowed inst['name', a:b].
        """

        if isinstance(key, tuple):
	    # support slicing	
            return self.data.ix[key[0], key[1]]
	else:
	    return self.data[key]

    def __setitem__(self,key,new):
        """Add data to inst.data dataFrame. inst['name'] = newData."""
        if isinstance(new, dict):
            # metadata should be included in dict
            self.data[key] = new.pop('data')
            # pass the rest to meta
            self.meta[key] = new
        else:
            if isinstance(key, tuple):
                self.data.ix[key[0],key[1]] = new
                if key[1] not in self.meta.data.index:
                    self.meta[key[1]] = {'long_name':key, 'units':''}
            elif isinstance(key, str):
                self.data[key] = new  
                if key not in self.meta.data.index:   
                    # add in default metadata because none was supplied
                    self.meta[key] = {'long_name':key, 'units':''} 
            elif isinstance(new, pds.DataFrame):
                self.data[key] = new[key]
                for ke in key:
                    if ke not in self.meta.data.index:   
                        # add in default metadata because none was supplied
                        self.meta[ke] = {'long_name':ke, 'units':''}                 
            else:
                raise ValueError("No support for supplied input key")           
    
    def copy(self):
        """Deep copy of the entire satellite object."""
        return copy.deepcopy(self)
    
    def _pass_func(*args, **kwargs):
        pass     

    def _assign_funcs(self, by_name=False, inst_module=None):        
        """Assign all external science instrument methods to Instrument object."""
        import importlib
        # set defaults
        self._list_rtn = self._pass_func
        self._load_rtn = self._pass_func
        self._default_rtn = self._pass_func
        self._clean_rtn = self._pass_func
        self._init_rtn = self._pass_func
        self._download_rtn = self._pass_func
            
        if by_name: 
            # look for code with filename name, any errors passed up
            inst = importlib.import_module(''.join(('.',self.platform,'_',
                        self.name)),  package='pysat.instruments')
 	elif inst_module is not None:
 	    # user supplied an object with relevant instrument routines
 	    inst = inst_module
 	else:
 	    # no module or name info, default pass functions assigned
 	    return
 	               
        try:
            self._load_rtn = inst.load
            self._list_rtn = inst.list_files
            self._download_rtn = inst.download
        except AttributeError:
            raise AttributeError(string.join(('A load, file_list, and download routine',
                    'are required for every instrument.'))) 
        try:
            self._default_rtn = inst.default
        except AttributeError:
            pass
        try:
            self._init_rtn = inst.init
        except AttributeError:
            pass   
        try:
            self._clean_rtn = inst.clean
        except AttributeError:
            pass

        return

    def _load_data(self, date=None, fid=None):
        """
        Load data for an instrument on given date or fid, dependng upon input.
        """

	if fid is not None:
	    # get filename based off of index value
	    # remove direct access of data
	    fname = self.files.files.iloc[fid]
	elif date is not None:
	    fname = self.files[date : date+pds.DateOffset(days=1)]
	else:
	    raise ValueError('Must supply either a date or file id number.')
   
        if len(fname) > 0:    
            load_fname = [os.path.join(self.files.data_path, f) for f in fname]
            data, mdata = self._load_rtn(load_fname, tag = self.tag) 
        else:
            data = pds.DataFrame(None)
            mdata = _meta.Meta()

        if not data.empty: 
            if not isinstance(data, pds.DataFrame):
                raise TypeError(string.join(('Data returned by instrument load',
                                'routine must be a pandas.DataFrame')))
            if not isinstance(mdata, _meta.Meta):
                raise TypeError('Metadata returned must be a pysat.Meta object')
            if date is not None:
                print string.join(('Returning',self.platform,self.name,self.tag,'data for', 
                              date.strftime('%D')))
            else:
                if len(fname) == 1:
                    # this check was zero
                    print string.join(('Returning',self.platform,self.name,self.tag,
                                        'data from',fname[0]))
                else:
                    print string.join(('Returning',self.platform,self.name,self.tag,
                            'data from',fname[0], ' :: ',fname[-1]))
        else:
            # no data signal
            print string.join(('No',self.platform,self.name,self.tag,'data for', 
                                date.strftime('%D')))
    
        return data, mdata
        
    def _load_next(self):
        '''Load the next days data (or file) without incrementing the date.
        Repeated calls will not advance date/file and will produce the same data
        
        Uses info stored in object to either increment the date, 
        or the file. Looks for self._load_by_date flag.  
	'''
	if self._load_by_date:
	    next_date = self.date + pds.DateOffset(days=1)
       	    return self._load_data(date=next_date)
       	else:
       	    return self._load_data(fid = self._fid+1)

    def _load_prev(self):
        """Load the next days data (or file) without decrementing the date.
        Repeated calls will not decrement date/file and will produce the same data
        
        Uses info stored in object to either decrement the date, 
        or the file. Looks for self._load_by_date flag.  
	"""
	if self._load_by_date:
	    prev_date = self.date - pds.DateOffset(days=1)
	    return self._load_data(date=prev_date)
	else:
	    return self._load_data(fid = self._fid-1)

    def load(self, yr=None, doy=None, date=None, fname=None, fid=None, 
                verifyPad=False, *arg, **kwargs):
        """
        Load instrument data into satellite object as .data

        Keywords:
            yr: year
            doy: day of year
            date: datetime object for date
            fname: filename to be loaded
            verifyPad: boolean to verify data padding is being applied (debug purposes)
        Summary:
            Loads data for a chosen instrument into .data. Any functions chosen
            by the user and added to the custom processing queue (.custom.add)
            are automatically applied to the data before it is available to 
            user in .data.
        """

        if date is not None:
            # date supplied getyrdoy checks if it is datetime
            year, doy = utils.getyrdoy(date)
            self.yr = year 
            self.doy = doy 
            self.date = date
            self._fid = None
            self._load_by_date = True
            inc = pds.DateOffset(days=1)
            curr = date
        elif (yr is not None) & (doy is not None):
            # if date not defined but both yr and doy are
            self.date = (pds.datetime(yr,1,1) ) + pds.DateOffset(days=(doy-1))
            self.yr = yr
            self.doy = doy
            self._fid = None
            self._load_by_date = True
            inc = pds.DateOffset(days=1)
            curr = self.date
        elif fname != None:
	    # date will have to be set later by looking at the data
            self.date = None
            self.yr = None
            self.doy = None
            self._load_by_date = False
            # if no index, called func tries to find file in instrument dir,
            # throws IOError if it fails
            self._fid = self.files.get_index(fname)
            inc = 1
            curr = self._fid.copy()
	elif fid != None:
            self._load_by_date = False	    
	    self._fid = fid
	    self.date = None
            self.yr = None
            self.doy = None
	    inc = 1
	    curr = fid
        else:
            raise TypeError('Must supply a yr,doy pair, or datetime object, '+ 
				'or filename to load data from.')
				
        self.orbits._reset()        
        # if pad is true, need to have a three day/file load
        if self.pad:
            if self._next_data.empty & self._prev_data.empty:
                # data has not already been loaded for previous and next days
                # load data for all three
                print 'Initializing three day/file window'
                # using current date or fid
                self._prev_data, self._prev_meta = self._load_prev()
                self._curr_data, self._curr_meta = self._load_data(date=self.date, fid=self._fid)
                self._next_data, self._next_meta = self._load_next()
            else:
                # moving forward in time
                if self._next_data_track == curr:
                    self._prev_data, self._prev_meta = self._curr_data, self._curr_meta
                    self._curr_data, self._curr_meta = self._next_data, self._next_meta
                    self._next_data, self._next_meta = self._load_next()
                # moving backward in time
                elif self._prev_data_track == curr:
                    self._next_data, self._next_meta = self._curr_data, self._curr_meta
                    self._curr_data, self._curr_meta = self._prev_data, self._prev_meta
                    self._prev_data, self._prev_meta = self._load_prev()
                # jumped in time/or switched from filebased to date based access
                else:
                    self._prev_data, self._prev_meta = self._load_prev()
                    self._curr_data, self._curr_meta = self._load_data(date=self.date, fid=self._fid)   
                    self._next_data, self._next_meta = self._load_next()

            # make tracking indexes consistent with new loads
            self._next_data_track = curr + inc
            self._prev_data_track = curr - inc
	    # attach data to object
	    if not self._curr_data.empty :
	        self.data = self._curr_data.copy()
	        self.meta = self._curr_meta.copy()
	    else:
	        self.data = pds.DataFrame(None)	
	        self.meta = _meta.Meta()
            # pad data based upon passed parameter
            offs = pds.DateOffset( **(self.padkws) )
            if (not self._prev_data.empty) & (not self.data.empty) :
                padLeft = self._prev_data[(self._curr_data.index[0]-offs):self._curr_data.index[0]]
                self.data = pds.concat([padLeft[0:-1], self.data])
            if  (not self._next_data.empty) & (not self.data.empty) :
                padRight = self._next_data[self._curr_data.index[-1]:(self._curr_data.index[-1]+offs)]
                self.data = pds.concat([self.data, padRight[1:]])
        # if self.pad is False, load single day
        else:
            self.data, self.meta  = self._load_data(date=self.date, fid=self._fid)       
        # check if load routine actually returns meta
        if self.meta.data.empty:
            self.meta[self.data.columns] = {'long_name':self.data.columns,
                                            'units':['']*len(self.data.columns)}
        if not self.data.empty:
            self._default_rtn(self)
	# clean
        if  (not self.data.empty) & (self.clean_level != 'none') :
            self._clean_rtn(self)   
        # apply custom functions
        if not self.data.empty:
            self.custom._apply_all(self)
        # remove the excess padding, if any applied
        if self.pad & (not self.data.empty) & (not verifyPad) :
            self.data = self.data[self._curr_data.index[0]:self._curr_data.index[-1]]
        # if loading by file set the yr, doy, and date
	if not self._load_by_date:
	    temp = self.data.index[0]
	    temp = pds.datetime(temp.year, temp.month, temp.day)
	    self.date = temp
	    self.yr, self.doy = utils.getyrdoy(self.date)
	                    
        sys.stdout.flush()
        return

    def download(self, start, stop, user=None, password=None):
        '''Download data for given Instrument object.'''
        import errno
        # make sure directories are there, otherwise create them
        try:
            os.mkdir(os.path.join(data_dir, self.platform))
        except OSError as exception:
            if exception.errno != errno.EEXIST:
                raise    
        try:
            os.mkdir(os.path.join(data_dir, self.platform, self.name))
        except OSError as exception:
            if exception.errno != errno.EEXIST:
                raise    
        try:
            os.mkdir(self.files.data_path)
        except OSError as exception:
            if exception.errno != errno.EEXIST:
                raise
        date_array = utils.season_date_range(start,stop)
        if user is None:
            self._download_rtn(date_array, tag=self.tag, 
                    data_path=self.files.data_path)
        else:
            self._download_rtn(date_array, tag=self.tag, 
                    data_path=self.files.data_path,user=user, password=password)	
        print 'Updating pysat file list'
        self.files.refresh(store=True)
        
    @property
    def bounds(self):
	"""
	Boundaries for iterating over instrument object by date or file.
	
	input: 
	   start: datetime object, filename, or None (default)
		  Defines start of iteration, if None uses first data date.
	   end:  datetime object, filename, or None (default)
		 Defines end of iteration, inclusive. If None uses last
		 data date. 	
        """
        return (self._iter_start, self._iter_stop)
    
    @bounds.setter        
    def bounds(self, value=None):
        #start=None, end=None):
	"""
	Set boundaries for iterating over instrument object by date or file.
	
	input: 
	   start: datetime object, filename, or None (default)
		  Defines start of iteration, if None uses first data date.
	   end:  datetime object, filename, or None (default)
		 Defines end of iteration, inclusive. If None uses last
		 data date. 	
        """
        if value is None:
            value = (None,None)
        if len(value) < 2:
            raise ValueError('Must supply both a start and end date/file'+
                  'Supply None if you want the first/last possible')

        start = value[0]
        end = value[1]

        if (start is None) and (end is None):
	    #set default
	    self._iter_start = [self.files.start_date]
	    self._iter_stop = [self.files.stop_date]
            self._iter_type = 'date'
            if self._iter_start[0] is not None:
                # check here in case Instrument is initialized with no input
                self._iter_list = utils.season_date_range(self._iter_start, self._iter_stop)

	elif hasattr(start, '__iter__') and hasattr(end, '__iter__'):
            base = type(start[0])
	    for s,t in zip(start, end):
	        if (type(s) != type(t)) or (type(s) != base):
	            raise ValueError('Start and end items must all be of the same type')
	    if isinstance(start[0], str):
                self._iter_type = 'file'
		self._iter_list = self.files.get_file_array(start, end)	        
            elif isinstance(start[0], pds.datetime):
                self._iter_type = 'date'
                self._iter_list = utils.season_date_range(start, end)
            else:
                raise ValueError ('Input is not a known type, string or datetime')
            self._iter_start = start
            self._iter_stop = end

        elif hasattr(start, '__iter__') or hasattr(end, '__iter__'):
            raise ValueError('Both start and end must be iterable if one bound is iterable')

        elif isinstance(start, str) or isinstance(end, str):
            if isinstance(start, pds.datetime) or isinstance(end, pds.datetime):
                raise ValueError('Not allowed to mix file and date bounds')
	    if start is None:
	        start = self.files.files[0]
	    if end is None:
	        end = self.file.files[-1]	
            self._iter_start = [start]
            self._iter_stop = [end]
            self._iter_list = self.files.get_file_array(self._iter_start,self._iter_stop)
            self._iter_type = 'file'

        elif isinstance(start, pds.datetime) or isinstance(end, pds.datetime):
	    if start is None:
	        start = self.files.start_date
	    if end is None:
	        end = self.files.stop_date
	    self._iter_start = [start]
	    self._iter_stop = [end]	
            self._iter_list = utils.season_date_range(start, end)
            self._iter_type = 'date'
	else:
	    raise ValueError('Provided an invalid combination of bounds. '+
	    'if specifying by file, both bounds must be by file. Other '+
	    'combinations of datetime objects and None are allowed.')


    def __iter__(self):
        """
        Iterates by loading subsequent days or files as appropriate.
        	
	Limits of iteration set by calling setBounds method.	
        """
        if self._iter_type == 'file':
            for fid in self._iter_list:
                self.load(fid=fid)
                yield self       
 
	elif self._iter_type == 'date': 
            for date in self._iter_list:
                self.load(date=date)
                yield self            
                
    def next(self):
        """
        Manually iterate through the data loaded in satellite object.
        
        Bounds of iteration and iteration type (day/file) are set by setBounds
        If there were no previous calls to load then the first
        day (default)/file will be loaded.
        """
        
	if self._iter_type == 'date':
       	    if self.date is not None:
       	        idx, = np.where(self._iter_list == self.date )
       	        if (len(idx)==0) | (idx+1 >= len(self._iter_list) ):
       	            raise StopIteration('Outside the set date boundaries.')
       	        else:
       	            idx += 1
       	            self.load(date = self._iter_list[idx[0]])
            else:
                self.load(date = self._iter_list[0])
                
        elif self._iter_type == 'file':
            if self._fid is not None:
       	        idx, = np.where(self._iter_list == self._fid )
       	        if (len(idx)==0) | (idx+1 >= len(self._iter_list) ):
       	            raise StopIteration('Outside the set file boundaries.')
       	        else:
       	            idx += 1
       	            self.load(fid = self._iter_list[idx[0]])
            else:
                self.load(fid = self._iter_list[0])

    def prev(self):
        """
        Manually iterate backwards through the data loaded in satellite object.
        
        Bounds of iteration and iteration type (day/file) are set by setBounds
        If there were no previous calls to load then the last
        day (default)/file will be loaded.
        """
        
	if self._iter_type == 'date':
       	    if self.date is not None:
       	        idx, = np.where(self._iter_list == self.date )
       	        if (len(idx)==0) | (idx-1 < 0 ):
       	            raise StopIteration('Outside the set date boundaries.')
       	        else:
       	            idx -= 1
       	            self.load(date = self._iter_list[idx[0]])
            else:
                self.load(date = self._iter_list[-1])
                
        elif self._iter_type == 'file':
            if self._fid is not None:
       	        idx, = np.where(self._iter_list == self._fid )
       	        if (len(idx)==0) | (idx-1 < 0 ):
       	            raise StopIteration('Outside the set file boundaries.')
       	        else:
       	            idx -= 1
       	            self.load(fid = self._iter_list[idx[0]])
            else:
                #last file by default
                self.load(fid = self._iter_list[-1])


    def to_netcdf3(self, fname=None):
        """
        Stores loaded data into a netCDF3 64-bit file.
        
        Stores 1-D data along dimension 'time' - the date time index.
        Stores object data (dataframes within dataframe) separately:
            The name of the object data is used to prepend extra variable
              dimensions within netCDF, key_2, key_3, first dimension time
            The index organizing the data stored as key_sample_index
            from_netcdf3 uses this naming scheme to reconstruct data structure
        The datetime index is stored as 'UNIX time'. netCDF-3 doesn't support
          64-bit integers so it is stored as a 64-bit float. This results in a
          loss of datetime precision when converted back to datetime index
          up to hundreds of nanoseconds. Use netCDF4 if this is a problem.
          
        All attributes attached to instrument meta are written to netCDF attrs.
        
        """
        
        import netCDF4
        
        with netCDF4.Dataset('/Users/rstoneba/Desktop/test.nc', 
                            mode='w', format='NETCDF3_64BIT') as out_data:
        
            num = len(self.data.index)
            out_data.createDimension('time', num)
            
            # write out the datetime index
            cdfkey = out_data.createVariable('unix_time', 'f8', dimensions=('time'),)
            cdfkey.units = 's'
            cdfkey.long_name = 'UNIX time'
            cdfkey.calendar = 'noleap'
            cdfkey[:] = self.data.index.astype(int)*1.E-9
            # store all of the data names in meta            
            for key in self.meta.data.index:
                if self[key].dtype != np.dtype('O'):
                    # not an object, simple column of data, write it out
                    cdfkey = out_data.createVariable(key, self[key].dtype, dimensions=('time'), )
                    cdfkey.units = self.meta[key].units
                    cdfkey.long_name = self.meta[key].long_name
                    cdfkey[:] = self[key].values
                else:
                    # we are dealing with a more complicated object
                    # presuming a series with a dataframe in each location
                    dims = np.shape(self[key].iloc[0])
                    obj_dim_names=[]
                    # don't need to recreate last dimension, 
                    # it covers number of columns
                    for i,dim in enumerate(dims[:-1]):
                        obj_dim_names.append(key+'_dim_%i' % (i+1))
                        out_data.createDimension(obj_dim_names[-1], dim)
                    var_dim = tuple(['time']+obj_dim_names)
                    
                    # iterate over columns and store
                    for col in self[key].iloc[0].columns:
                        coltype = self[key].iloc[0][col].dtype
                        if coltype == np.int64 :
                            coltype = np.int32
                        cdfkey = out_data.createVariable(col, coltype, dimensions=var_dim)
                        cdfkey.long_name = col
                        cdfkey.units = ''
                        for i in xrange(num):
                            cdfkey[i,:] = self[key].iloc[i][col].values.astype(coltype)
                            
                    # store the dataframe index for each time of main dataframe
                    datetime_flag = False
                    coltype = self[key].iloc[0].index.dtype
                    # check for datetime index
                    if coltype == np.dtype('<M8[ns]'):
                        coltype = 'f8' #self[key].iloc[0].index.to_native_types()[0].dtype
                        datetime_flag = True
                    
                    cdfkey = out_data.createVariable(key+'_sample_index', 
                                        coltype, dimensions=var_dim)
                    if datetime_flag:
                        cdfkey.units='s'
                        cdfkey.long_name = 'UNIX time'
                        for i in xrange(num):
                            cdfkey[i,:] = self[key].iloc[i].index.astype(int)*1.E-9
                    else:
                        cdfkey.units=''
                        cdfkey.long_name = key
                        for i in xrange(num):
                            cdfkey[i,:] = self[key].iloc[i].index.to_native_types()

                    if self[key].iloc[0].index.name is not None:
                        cdfkey.long_name = self[key].iloc[0].index.name
                    
            # store any non standard attributes
            base_attrb = dir(Instrument())
            this_attrb = dir(self)
            
            adict = {}
            for key in this_attrb:
                if key not in base_attrb:
                    if key[0] != '_':
                        adict[key] = self.__getattribute__(key)
            # store any non-standard attributes attached to meta
            base_attrb = dir(_meta.Meta())
            this_attrb = dir(self.meta)
            for key in this_attrb:
                if key not in base_attrb:
                    if key[0] != '_':
                        adict[key] = self.__getattribute__(key)
            adict['pysat_version'] = 1.0
            adict['Conventions'] = 'CF-1.6'
            out_data.setncatts(adict)


